/* ---------------------------------------------------------------------------
   Based on
   Bare AsciiDoc styles
   Ryan Tomayko <r@tomayko.com>

   Heavily modified by
   Kristoffer Gronlund <kgronlund@suse.com>
   --------------------------------------------------------------------------- */

/* Fonts */

body {
    font-family:'Open Sans', 'lucida grande',verdana,helvetica,arial,sans-serif;
    font-size: 16px;
    line-height: 21px;
}

.small {
  font-size: 12px; /* 75% of the baseline */
}

.large {
  font-size: 20px; /* 125% of the baseline */
}

em {
    font-style:italic;
}

strong {
    font-weight:bold;
}

.monospaced {
    font-family: 'Menlo', 'Monaco', 'Consolas', 'lucida console', 'bitstream vera sans mono', 'courier new', monospace;
}

dt {
    font-weight:normal;
}

h1, h2, h3, h4, h5 {
    font-family:'Open Sans', 'lucida grande',verdana,helvetica,arial,sans-serif;
    font-weight:normal;
}

h1 {
    font-size:2.6rem;
    line-height:1.428;
}

h2 {
    font-size:2rem;
    line-height:1.36363636; /* repeating, of course */
}

h3 {
    font-size:1.6rem;
    line-height:1.1;
}

h4 {
    font-weight: bold;
    font-size:1.3rem;
    line-height:1.538;
}

h5 {
    font-size:1.2rem;
    font-style:italic;
    line-height:1.538;
}

pre {
    font-family:'Menlo', 'Monaco', consolas, 'lucida console', 'bitstream vera sans mono', 'courier new', monospace;
    font-size: 14px;
}

#header h1 {
    font-size: 42px;
    margin: 0px;
    display: inline;
}

#topbar {
    display: inline;
    font-size: 18px;
}

#topbar-small {
    display: none;
    font-size: 14px;
}



/* Style */

body {
    margin: 0px 0px;
    padding: 0px;
    width: 100%;
    color:#333;
    background: #fff;
}

.monospaced {
    color: #211;
    background-color: #fafaf8;
}

p {
    margin-bottom: 1.3636rem;
}

ul, ol, dl {
    margin-top: 1rem;
    margin-bottom: 2rem;
}

ul p {
    margin: 10px 0;
}

dl {
    margin-left:40px
}

dt {
    color:#000;
}

h1, h2, h3, h4, h5 {
    color:#000;
}

h2, h3, h4, h5 {
    padding-bottom: 0.333rem;
    border-bottom: 1px solid #eee;
}

h1 {
    margin:0px;
    margin-top: 48px;
}

h2 {
    margin-top: 36px;
    margin-bottom: 1.5rem;
}

h3 {
    margin: 0px;
    margin-top: 30px;
}

pre {
    color: #211;
    overflow-x: auto;
}

#header {
    background: #ecf0f1;
    padding-left: 24px;
    padding-top: 4px;
    padding-bottom: 0px;
    border-bottom: 2px solid #efefea;
    width: 100%;
    height: 86px;
    overflow: hidden;
}

#header a {
    text-decoration: none;
    color: #34495e;
}

#header a:hover {
    color:#ee3300;
}

#topbar ul {
    list-style: none;
    display: inline;
}

#topbar li {
    list-style: none;
    display: inline;
    padding-right: 1rem;
}

#topbar-small ul {
    list-style: none;
    display: inline;
    padding-left: 0.5rem;
}

#topbar-small li {
    list-style: none;
    display: inline;
    padding-right: 0.5rem;
}


#container {
    max-width: 720px;
    margin-left: 240px;
    padding-left: 8px;
    text-align:left;
}

#author {
    color:#999;
}

a {
    text-decoration: none;
    color:#419eda;
}

a:active {
 color:#6ec654;
}

a:hover {
 color:#ee3300;
 text-decoration: underline;
}


#content {
}

h1 {
    margin-left: auto;
    margin-right: auto;
    width: 551px;
    text-align: center;
    margin-bottom: 1.5rem;
}

.frontpage-image {
    margin-left: auto;
    margin-right: auto;
    width: 551px;
}

.title, .sidebar-title {
    font-weight:normal;
    color:#000;
    margin-bottom:0;
}

div.content {
    margin: 8px;
    padding: 0;
}

div.admonitionblock .title {
    font-weight:bold;
}

div.admonitionblock {
    margin:30px 0px;
    color:#555;
}

div.admonitionblock td.icon {
    width:30px;
    padding-right:20px;
    padding-left:20px;
    text-transform:uppercase;
    font-weight:bold;
    color:#888;
}

div.listingblock .content {
    border-left:4px solid #419eda;
    padding:8px;

    background: #faf7f8;

    background-image: -moz-linear-gradient(left, right,
                                           from(#faf7f8),
                                           to(#ffffff));

    background-image: -webkit-gradient(linear, left top, right bottom,
                                       color-stop(0.00, #faf7f8),
                                       color-stop(1.00, #ffffff));

}

div.listingblock .content pre {
    margin:0;
}

div.literalblock .content {
    margin-left: 20px;
}

div.verseblock .content {
    white-space:pre
}

div.sidebarblock {
    margin-top: 1.5rem;
    margin-bottom: 2rem;
}

div.sidebarblock > div.content {
    border-left:4px solid #ee3300;
    background: #faf7f8;
    padding:0 10px;
    color:#222;
    font-size: 14px;
    line-height:18px;
    max-width: 720px;

    background-image: -moz-linear-gradient(left, right,
                                           from(#faf7f8),
                                           to(#ffffff));

    background-image: -webkit-gradient(linear, left top, right bottom,
                                       color-stop(0.00, #faf7f8),
                                       color-stop(1.00, #ffffff));

}

div.sidebarblock .title {
    margin:10px 0;
    font-weight:bold;
    font-size: 14px;
    color:#442;
}

.quoteblock-content {
    font-style:italic;
    color:#444;
    margin-left:40px;
}

.quoteblock-content .attribution {
    font-style:normal;
    text-align:right;
    color:#000;
}

.exampleblock-content *:first-child { margin-top:0 }
.exampleblock-content {
    border-left:2px solid silver;
    padding-left:8px;
}

#footnotes {
    text-align:left;
}

#footnotes hr {
    height: 1px;
    color: #ccc;
    width: 80%;
}

#footer {
    font-size: 12px;
    color:#888;
    margin-top:40px;
    text-align: right;
}

.nav {
    margin-bottom: 0;
    padding-left: 0;
    list-style: none;
}

.nav li {
    line-height: 4rem;
}

.nav a {
    font-size: 20px;
    text-decoration: none;
}

.feedEkList .newsItem {
    list-style-type: none;
}

.feedEkList .itemTitle {
    font-size: large;
}

.feedEkList .itemDate {
    font-size: smaller;
}

.feedEkList .itemContent {
}

@media screen {
    #toc {
        position: fixed;
        top: 120px;
        left: 4px;
        margin: 0px;
        font-size: 12px;
        line-height: 1.2em;
    }

    #toc a .monospaced {
        color:#419eda;
    }

    #toc a {
        text-decoration: none;
    }

    #toc .toclevel1 {
        padding: 1px;
        margin-top: 8px;
        font-size: 14px;
        line-height: 16px;
    }

    #toc .toclevel2 {
        margin-left: 8px;
        padding-left: 4px;
        font-size: 12px;
        line-height: 16px;
    }

    #toc .toclevel3 {
        margin-left: 24px;
        padding-left: 4px;
        font-size: 11px;
        line-height: 15px;
        font-weight: bold;
    }

    #toctitle {
        margin:20px 0;

    }
}

@media screen and (max-width: 900px) {
    #toc {
        display: none;
    }

    #container {
        max-width: 720px;
        margin: 0px auto;
        text-align:left;
        padding-right: 8px;
    }

    #topbar {
        display: none;
    }

    #topbar-small {
        display: inline;
    }
}

@media screen and (min-width: 900px) and (max-width: 1280px) {
    #toc {
        position: absolute;
        overflow: hidden;
        top: 120px;
        left: 4px;
        max-width: 240px;
    }
}

@media screen and (min-width: 1280px) {
    #toc {
        position: absolute;
        overflow: hidden;
        top: 120px;
        left: 4px;
        max-width: 240px;
    }

    #container {
        max-width: 960px;
    }
}

@media screen and (min-width: 1500px) {
    #container {
        margin-left: auto;
        margin-right: auto;
    }
}

/* pygments highlighting */

.hll { background-color: #ffffcc }
.c { color: #999988; font-style: italic } /* Comment */
.err { color: #a61717; background-color: #e3d2d2 } /* Error */
.k { color: #000000; font-weight: bold } /* Keyword */
.o { color: #000000; font-weight: bold } /* Operator */
.cm { color: #999988; font-style: italic } /* Comment.Multiline */
.cp { color: #999999; font-weight: bold; font-style: italic } /* Comment.Preproc */
.c1 { color: #999988; font-style: italic } /* Comment.Single */
.cs { color: #999999; font-weight: bold; font-style: italic } /* Comment.Special */
.gd { color: #000000; background-color: #ffdddd } /* Generic.Deleted */
.ge { color: #000000; font-style: italic } /* Generic.Emph */
.gr { color: #aa0000 } /* Generic.Error */
.gh { color: #999999 } /* Generic.Heading */
.gi { color: #000000; background-color: #ddffdd } /* Generic.Inserted */
.go { color: #888888 } /* Generic.Output */
.gp { color: #555555 } /* Generic.Prompt */
.gs { font-weight: bold } /* Generic.Strong */
.gu { color: #aaaaaa } /* Generic.Subheading */
.gt { color: #aa0000 } /* Generic.Traceback */
.kc { color: #000000; font-weight: bold } /* Keyword.Constant */
.kd { color: #000000; font-weight: bold } /* Keyword.Declaration */
.kn { color: #000000; font-weight: bold } /* Keyword.Namespace */
.kp { color: #000000; font-weight: bold } /* Keyword.Pseudo */
.kr { color: #000000; font-weight: bold } /* Keyword.Reserved */
.kt { color: #445588; font-weight: bold } /* Keyword.Type */
.m { color: #009999 } /* Literal.Number */
.s { color: #d01040 } /* Literal.String */
.na { color: #008080 } /* Name.Attribute */
.nb { color: #0086B3 } /* Name.Builtin */
.nc { color: #445588; font-weight: bold } /* Name.Class */
.no { color: #008080 } /* Name.Constant */
.nd { color: #3c5d5d; font-weight: bold } /* Name.Decorator */
.ni { color: #800080 } /* Name.Entity */
.ne { color: #990000; font-weight: bold } /* Name.Exception */
.nf { color: #990000; font-weight: bold } /* Name.Function */
.nl { color: #990000; font-weight: bold } /* Name.Label */
.nn { color: #555555 } /* Name.Namespace */
.nt { color: #000080 } /* Name.Tag */
.nv { color: #008080 } /* Name.Variable */
.ow { color: #000000; font-weight: bold } /* Operator.Word */
.w { color: #bbbbbb } /* Text.Whitespace */
.mf { color: #009999 } /* Literal.Number.Float */
.mh { color: #009999 } /* Literal.Number.Hex */
.mi { color: #009999 } /* Literal.Number.Integer */
.mo { color: #009999 } /* Literal.Number.Oct */
.sb { color: #d01040 } /* Literal.String.Backtick */
.sc { color: #d01040 } /* Literal.String.Char */
.sd { color: #d01040 } /* Literal.String.Doc */
.s2 { color: #d01040 } /* Literal.String.Double */
.se { color: #d01040 } /* Literal.String.Escape */
.sh { color: #d01040 } /* Literal.String.Heredoc */
.si { color: #d01040 } /* Literal.String.Interpol */
.sx { color: #d01040 } /* Literal.String.Other */
.sr { color: #009926 } /* Literal.String.Regex */
.s1 { color: #d01040 } /* Literal.String.Single */
.ss { color: #990073 } /* Literal.String.Symbol */
.bp { color: #999999 } /* Name.Builtin.Pseudo */
.vc { color: #008080 } /* Name.Variable.Class */
.vg { color: #008080 } /* Name.Variable.Global */
.vi { color: #008080 } /* Name.Variable.Instance */
.il { color: #009999 } /* Literal.Number.Integer.Long */
.highlight .-Color-Black { color: #000000 } /* Color.Black */
.highlight .-Color-Blue { color: #0000c0 } /* Color.Blue */
.highlight .-Color-Cyan { color: #008080 } /* Color.Cyan */
.highlight .-Color-Green { color: #008000 } /* Color.Green */
.highlight .-Color-Magenta { color: #c000c0 } /* Color.Magenta */
.highlight .-Color-Red { color: #c00000 } /* Color.Red */
.highlight .-Color-White { color: #c0c0c0 } /* Color.White */
.highlight .-Color-Yellow { color: #808000 } /* Color.Yellow */
